#!/usr/bin/env python
#
#  bdict.py
"""
Provides bdict, a dictionary where keys and values are also stored the other way round.
"""
#
#  Copyright © 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#  Improved May 2020 with suggestions from
#      https://treyhunner.com/2019/04/why-you-shouldnt-inherit-from-list-and-dict-in-python/
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from collections import UserDict
from typing import AbstractSet, Any, Iterable, Optional, Tuple, TypeVar, Union, ValuesView, overload

# this package
from cawdrey.base import KT, VT, T

__all__ = ["bdict"]


class bdict(UserDict):
	"""
	Returns a new dictionary initialized from an optional positional argument,
	and a possibly empty set of keyword arguments.

	Each ``key: value`` pair is entered into the dictionary in both directions,
	so you can perform lookups with either the key or the value.

	If no positional argument is given, an empty dictionary is created.

	If a positional argument is given and it is a mapping object, a dictionary
	is created with the same key-value pairs as the mapping object.
	Otherwise, the positional argument must be an iterable object.
	Each item in the iterable must itself be an iterable with exactly two
	objects. The first object of each item becomes a key in the new
	dictionary, and the second object the corresponding value.

	If keyword arguments are given, the keyword arguments and their values are
	added to the dictionary created from the positional argument.

	If an attempt is made to add a key or value that already exists in the
	dictionary a :exc:`ValueError` will be raised.

	Keys or values of :py:obj:`None`, :py:obj:`True` and :py:obj:`False` will be stored internally as
	``"_None"``, ``"_True"`` and ``"_False"`` respectively
	"""  # noqa: D400

	# Based on https://stackoverflow.com/a/1063393 by https://stackoverflow.com/users/9493/brian

	# Improved May 2020 with suggestions from
	# https://treyhunner.com/2019/04/why-you-shouldnt-inherit-from-list-and-dict-in-python/

	def __init__(self, seq: Optional[Iterable] = None, **kwargs):
		# if seq and kwargs:
		# 	raise TypeError(f'expected at most 1 arguments, got {len(kwargs)-1:d}')

		super().__init__(seq, **kwargs)
		# if seq:
		# 	for key, value in dict(seq).items():
		# 		self.__setitem__(key, value)
		# else:
		# 	for key, value in kwargs.items():
		# 		self.__setitem__(key, value)

	def __setitem__(self, key, val):
		"""
		Set ``self[key]`` to ``value``.

		:param key:
		:param val:
		"""

		if key in self:
			del self[self[key]]
		if val in self:
			del self[val]
		#
		# if key in self or val in self:
		# 	if key in self and self[key] != val:
		# 		raise ValueError(f"The key '{key}' is already present in the dictionary")
		# 	if val in self and self[val] != key:
		# 		raise ValueError(f"The key '{val}' is already present in the dictionary")

		if key is None:
			key = "_None"
		if val is None:
			val = "_None"

		if isinstance(key, bool):
			if key:
				key = "_True"
			else:
				key = "_False"

		if isinstance(val, bool):
			if val:
				val = "_True"
			else:
				val = "_False"

		self.data[key] = val
		self.data[val] = key

	def __delitem__(self, key: KT):
		"""
		Delete ``self[key]``.

		:param key:
		"""

		value = self.data.pop(key)
		self.data.pop(value, None)

	def __getitem__(self, key: KT) -> VT:
		"""
		Return ``self[key]``.

		:param key:
		"""

		if key is None:
			key = "_None"

		if isinstance(key, bool):
			if key:
				key = "_True"
			else:
				key = "_False"

		val = super().__getitem__(key)

		if val == "_None":
			return None
		elif val == "_True":
			return True
		elif val == "_False":
			return False
		else:
			return val

	def __contains__(self, key: object) -> bool:
		"""
		Return ``key in self``.

		:param key:
		"""

		if key is None:
			key = "_None"

		if isinstance(key, bool):
			if key:
				key = "_True"
			else:
				key = "_False"

		return super().__contains__(key)

	@overload
	def get(self, k: KT) -> Optional[VT]:
		...  # pragma: no cover

	@overload
	def get(self, k: KT, default: Union[VT, T]) -> Union[VT, T]:
		...  # pragma: no cover

	def get(self, k, default=None):
		"""
		Return the value for ``k`` if ``k`` is in the dictionary, else ``default``.

		:param k: The key to return the value for.
		:param default: The value to return if ``key`` is not in the dictionary.
		"""

		return super().get(k, default)

	def items(self) -> AbstractSet[Tuple[KT, VT]]:
		r"""
		Returns a set-like object providing a view on the :class:`~.bdict`\'s items.
		"""

		return super().items()

	def keys(self) -> AbstractSet[KT]:
		r"""
		Returns a set-like object providing a view on the :class:`~.bdict`\'s keys.
		"""

		return super().keys()

	def values(self) -> ValuesView[VT]:
		r"""
		Returns an object providing a view on the :class:`~.bdict`\'s values.
		"""

		return super().values()

	def clear(self) -> None:
		"""
		Removes all items from the :class:`~.bdict`.
		"""

		return super().clear()



#!/usr/bin/env python
#
#  __init__.py
"""
Provides several useful custom dictionaries.
"""
#
#  Copyright © 2019-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#  Copyright © 2015 Warren Smith
#  Copyright © Marco Sulla
#  Copyright © 2012 Santiago Lezica
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# this package
from ._bdict import bdict
from ._frozendict import frozendict
from .alphadict import AlphaDict, alphabetical_dict
from .base import FrozenBase, MutableBase
from .frozenordereddict import FrozenOrderedDict
from .nonelessdict import NonelessDict, NonelessOrderedDict
from .tally import Tally

__author__: str = "Dominic Davis-Foster"
__copyright__: str = "2020 Dominic Davis-Foster"
__license__: str = "LGPLv3+"
__version__: str = "0.3.0"
__email__: str = "dominic@davis-foster.co.uk"

__all__ = [
		"__version__",
		"alphabetical_dict",
		"AlphaDict",
		"bdict",
		"FrozenBase",
		"frozendict",
		"FrozenOrderedDict",
		"MutableBase",
		"NonelessDict",
		"NonelessOrderedDict",
		"Tally",
		]



#!/usr/bin/env python
#
#  alphadict.py
"""
Provides AlphaDict, a frozen OrderedDict where the keys are stored alphabetically.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from collections import OrderedDict
from typing import Iterable, Optional

# 3rd party
from domdf_python_tools.doctools import prettify_docstrings

# this package
from .base import KT, VT
from .frozenordereddict import FrozenOrderedDict

__all__ = ["alphabetical_dict", "AlphaDict"]


def alphabetical_dict(**kwargs):
	"""
	Returns an :class:`~collections.OrderedDict` with the keys sorted alphabetically.

	:param kwargs:
	:type kwargs:

	:return:
	:rtype:
	"""

	return OrderedDict(sorted(kwargs.items()))


@prettify_docstrings
class AlphaDict(FrozenOrderedDict[KT, VT]):
	r"""
	Initialize an immutable, alphabetised dictionary.

	The signature is the same as regular dictionaries.

	| ``AlphaDict()`` -> new empty AlphaDict
	| ``AlphaDict(mapping)`` -> new AlphaDict initialized from a mapping object's (key, value) pairs
	| ``AlphaDict(iterable)`` -> new AlphaDict initialized as if via:

		.. code-block:: python

			d = {}
			for k, v in iterable:
				d[k] = v

	| ``AlphaDict(**kwargs)`` -> new AlphaDict initialized with the ``name=value`` pairs in the keyword argument list.

		For example:

		.. code-block::

			AlphaDict(one=1, two=2)
	"""

	def __init__(self, seq: Optional[Iterable] = None, **kwargs):
		if seq:
			kwargs = dict(seq)

		super().__init__(sorted(kwargs.items()))



#!/usr/bin/env python
#
#  frozenordereddict.py
"""
Provides FrozenOrderedDict, an immutable ordered dictionary.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#  Copyright © 2015 Warren Smith
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
import operator
from collections import OrderedDict
from functools import reduce
from typing import AbstractSet, Optional, Tuple, TypeVar, Union, ValuesView, overload

# this package
from .base import KT, VT, FrozenBase

__all__ = ["FrozenOrderedDict"]

T = TypeVar('T')


class FrozenOrderedDict(FrozenBase[KT, VT]):
	"""
	An immutable OrderedDict.
	It can be used as a drop-in replacement for dictionaries where immutability is desired.
	"""

	dict_cls = OrderedDict

	def __init__(self, *args, **kwargs):
		if hasattr(self, "_dict"):
			raise TypeError(f"`{self.__class__}` can only be initialised once.")

		super().__init__(*args, **kwargs)

	def copy(self, *args, **kwargs):
		"""
		Return a copy of the :class:`~cawdrey.frozenordereddict.FrozenOrderedDict`.

		:param args:
		:param kwargs:
		"""

		new_dict = self._dict.copy()

		if args or kwargs:
			new_dict.update(OrderedDict(*args, **kwargs))

		return self.__class__(new_dict)

	def __hash__(self) -> int:
		"""
		Return :func:`hash(self) <hash>`.
		"""

		if self._hash is None:
			self._hash = reduce(operator.xor, map(hash, self.items()), 0)

		return self._hash

	@overload
	def get(self, k: KT) -> Optional[VT]:
		...  # pragma: no cover

	@overload
	def get(self, k: KT, default: Union[VT, T]) -> Union[VT, T]:
		...  # pragma: no cover

	def get(self, k, default=None):
		"""
		Return the value for ``k`` if ``k`` is in the dictionary, else ``default``.

		:param k: The key to return the value for.
		:param default: The value to return if ``key`` is not in the dictionary.
		"""

		return super().get(k, default)

	def items(self) -> AbstractSet[Tuple[KT, VT]]:
		r"""
		Returns a set-like object providing a view on the :class:`~.FrozenOrderedDict`\'s items.
		"""

		return super().items()

	def keys(self) -> AbstractSet[KT]:
		r"""
		Returns a set-like object providing a view on the :class:`~.FrozenOrderedDict`\'s keys.
		"""

		return super().keys()

	def values(self) -> ValuesView[VT]:
		r"""
		Returns an object providing a view on the :class:`~.FrozenOrderedDict`\'s values.
		"""

		return super().values()

	def __contains__(self, key: object) -> bool:
		"""
		Return ``key in self``.

		:param key:
		"""

		return super().__contains__(key)

	def __getitem__(self, key: KT) -> VT:
		"""
		Return ``self[key]``.

		:param key:
		"""

		return super().__getitem__(key)



#!/usr/bin/env python
#
#  tally.py
"""
Subclass of :class:`collections.Counter` with additional methods.

.. versionadded:: 0.3.0
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  most_common method based on CPython.
#  Licensed under the Python Software Foundation License Version 2.
#  Copyright © 2001-2020 Python Software Foundation. All rights reserved.
#  Copyright © 2000 BeOpen.com. All rights reserved.
#  Copyright © 1995-2000 Corporation for National Research Initiatives. All rights reserved.
#  Copyright © 1991-1995 Stichting Mathematisch Centrum. All rights reserved.
#

# stdlib
import heapq
from numbers import Real
from operator import itemgetter
from typing import Counter, Dict, Iterable, List, Optional, Tuple, TypeVar, Union, overload

# 3rd party
from typing_extensions import Protocol, runtime_checkable

# this package
from cawdrey.base import KT

__all__ = ["_F", "SupportsMostCommon", "Tally", "Percentage"]

_F = TypeVar("_F", float, int, Real)


@runtime_checkable
class SupportsMostCommon(Protocol[KT]):
	"""
	:class:`typing.Protocol` for classes which support a :class:`collections.Counter`-like
	:meth:`collections.Counter.most_common` method.
	"""  # noqa: D400

	def items(self) -> Iterable[Tuple[KT, float]]:
		"""
		Returns an iterator over the mapping’s items (as ``(key, value)`` pairs).
		"""  # noqa: D402

		raise NotImplementedError

	def most_common(self, n: Optional[int] = None) -> Union[List[Tuple[KT, float]], List[Tuple[KT, int]]]:
		"""
		List the ``n`` most common elements and their counts from the most common to the least.
		If ``n`` is :py:obj:`None` then list all element counts.

		.. code-block:: python

			>>> Counter('abracadabra').most_common(3)
			[('a', 5), ('b', 2), ('r', 2)]

		:param n:
		"""

		raise NotImplementedError


class Tally(Counter[KT]):
	"""
	Subclass of :class:`collections.Counter` with additional methods.

	.. versionadded:: 0.3.0
	"""

	def as_percentage(self) -> "Percentage[KT]":
		"""
		Returns the :class:`~.Tally` as a :class:`collections.OrderedDict`
		comprising the count for each element as a percentage of the sum of all elements.

		.. important::
			The sum of the dictionary's values may not add up to exactly ``1.0``
			due to limitations of floating-point numbers.
		"""  # noqa: D400

		# Based on https://stackoverflow.com/a/48952797
		# By https://stackoverflow.com/users/271351/cjbarth
		# CC BY-SA 3.0

		total = self.total
		return Percentage((i, count / total) for i, count in self.most_common())

	@property
	def total(self) -> int:
		"""
		Returns the total count for all elements.
		"""

		return sum(self.values())

	@overload
	def get_percentage(self, item: KT) -> Optional[float]:
		...  # pragma: no cover

	@overload
	def get_percentage(self, item: KT, default: _F) -> Union[_F, float]:
		...  # pragma: no cover

	def get_percentage(self, item: KT, default: Optional[_F] = None) -> Union[None, _F, float]:
		"""
		Returns the count for ``item``, as a percentage of the sum of all elements.

		:param item:
		:param default: A default percentage (as a :class:`float`) to return if ``item`` is not in the dictionary.
		"""

		value = self.get(item)

		if value is None:
			return default
		else:
			return value / self.total

	def most_common(self, n: Optional[int] = None) -> List[Tuple[KT, int]]:
		"""
		List the ``n`` most common elements and their counts from the most common to the least.
		If ``n`` is :py:obj:`None` then list all element counts.

		.. code-block:: python

			>>> Tally('abracadabra').most_common(3)
			[('a', 5), ('b', 2), ('r', 2)]

		:param n:
		"""

		return super().most_common(n)


class Percentage(Dict[KT, float]):
	"""
	Provides a dictionary interface, but with :class:`collections.Counter`'s
	:meth:`collections.Counter.most_common` method.

	Represents the return type of :meth:`cawdrey.tally.Tally.as_percentage()`.
	"""  # noqa

	def most_common(self, n: Optional[int] = None) -> List[Tuple[KT, float]]:
		"""
		List the ``n`` most common elements and their counts from the most common to the least.
		If ``n`` is :py:obj:`None` then list all element counts.

		.. code-block:: python

			>>> Tally('abracadabra').as_percentage().most_common(3)
			[('a', 0.45454545454545453), ('b', 0.18181818181818182), ('r', 0.18181818181818182)]

		:param n:
		"""

		# Emulate Bag.sortedByCount from Smalltalk
		if n is None:
			return sorted(self.items(), key=itemgetter(1), reverse=True)

		return heapq.nlargest(n, self.items(), key=itemgetter(1))



#!/usr/bin/env python
#
#  nonelessdict.py
"""
Provides NonelessDict.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#  Copyright © Jeremy Mayeres
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
import operator
from collections import OrderedDict
from functools import reduce
from typing import Optional

# 3rd party
from domdf_python_tools.doctools import prettify_docstrings

# this package
from .base import KT, VT, MutableBase

__all__ = ["NonelessDict", "NonelessOrderedDict"]


@prettify_docstrings
class NonelessDict(MutableBase[KT, VT]):
	"""
	A wrapper around dict that will check if a value is
	:py:obj:`None`/empty/:py:obj:`False`, and not add the key in that case.

	Use the :meth:`~.NonelessDict.set_with_strict_none_check` method to check only
	for :py:obj:`None`.
	"""  # noqa: D400

	dict_cls = dict  # type: ignore

	def __init__(self, *args, **kwargs):
		if hasattr(self, "_dict"):
			raise TypeError(f"`{self.__class__}` can only be initialised once.")

		super().__init__(*args, **kwargs)

	def copy(self, **add_or_replace):
		"""
		Return a copy of the dictionary.
		"""

		return self.__class__(self, **add_or_replace)

	def __hash__(self) -> int:
		if self._hash is None:
			h = 0
			for key, value in self._dict.items():
				h ^= hash((key, value))
			self._hash = h
		return self._hash

	def set_with_strict_none_check(self, key: KT, value: Optional[VT]) -> None:
		"""
		Set ``key`` in the dictionary to ``value``, but skipping :py:obj:`None` values.

		:param key:
		:param value:
		"""

		if value is not None:
			self._dict[key] = value

	def __setitem__(self, key: KT, value: Optional[VT]):
		if value:
			return super().__setitem__(key, value)


@prettify_docstrings
class NonelessOrderedDict(MutableBase[KT, VT]):
	"""
	A wrapper around OrderedDict that will check if a value is None/empty/False,
	and not add the key in that case.
	Use the set_with_strict_none_check function to check only for None
	"""  # noqa: D400

	dict_cls = OrderedDict  # type: ignore

	def __init__(self, *args, **kwargs):
		if hasattr(self, "_dict"):
			raise TypeError(f"`{self.__class__}` can only be initialised once.")

		super().__init__(*args, **kwargs)

	def copy(self, *args, **kwargs):
		"""
		Return a copy of the dictionary.
		"""

		new_dict = self._dict.copy()

		if args or kwargs:
			new_dict.update(OrderedDict(*args, **kwargs))

		return self.__class__(new_dict)

	def __hash__(self) -> int:
		if self._hash is None:
			self._hash = reduce(operator.xor, map(hash, self.items()), 0)

		return self._hash

	def set_with_strict_none_check(self, key: KT, value: Optional[VT]) -> None:
		"""
		Set ``key`` in the dictionary to ``value``, but skipping :py:obj:`None` values.

		:param key:
		:param value:
		"""

		if value is not None:
			self._dict[key] = value

	def __setitem__(self, key: KT, value: Optional[VT]):
		if value:
			return super().__setitem__(key, value)
